
'''
@Modified:  2024-5-30
@Brief:     Update Image Safely 
'''
# !usr/bin/python3
import paho.mqtt.client as mqtt
import os
import threading
import json
import time
import sys
import zlib
#----------------------------------------------------------------------------------------#

# 测试平台
mqttConfig = {
    'mqttServer': 'www.witium.com.cn',
    'mqttPort': 12883,
    'userName': 'witcd',
    'password': 'Witium37774020'
}

# 正式平台
# mqttConfig = {
#     'mqttServer': '116.62.127.242',
#     'mqttPort': 1883,
#     'userName': 'witcd',
#     'password': 'Witium37774020'
# }



# # 颖泰平台
# mqttConfig = {
#     'mqttServer': '112.17.190.17',
#     'mqttPort': 1883,
#     'userName': 'witcd',
#     'password': 'Witium37774020'
# }

StampingLen = 14
StampingMark={'GWEV','GWAV','GWSV','SREV','SRAV','SRSV','TEST'}

dev_type = 'WTG906F'

updateCheckInTopic = 'WT/%s/Update/CheckIn' % dev_type
updateResponseTopic = 'WT/%s/Update/Response' % dev_type
updateTopic = 'WT/%s/{0}/Update' % dev_type



updateList = [{'sn': '21126112', 'package': 0, 'time': time.time(), 'send': True}]
OldImage = "at32_GatewayBinDir/v2.0.bin"
NewImage = "at32_GatewayBinDir/v2.0.bin"
cmd7Topic = 'WT/{}/{}/BackDoor'.format(dev_type,updateList[0]['sn'])
cmd7ResTopic = 'WT/{}/{}/BackDoorResp'.format(dev_type,updateList[0]['sn'])


CheckPassFlag = False


#----------------------------------------------------------------------------------------#

def get_Image_Size(file_path):
    file_size = 0
    with open(file_path, 'rb') as file:
        file_data = file.read()
        file_size = len(file_data)
    return file_size


def Stamping_To_Image(file_path, stamping):
    with open(file_path, 'ab') as file:
        file.write(stamping.encode())


def Last_Fourteen_Letter_Print(fiel_path):
    with open(file_path, 'rb') as file:
        file.seek(-StampingLen, 2)
        content = file.read()
        print("从倒数第{}个字符开始到结尾的内容：".format(StampingLen), content)
        return content

def Judge_Exit(content):
    for mark in StampingMark:
        if mark.encode('utf-8') in content:
            print("该镜像已经写入钢印！")
            return False
    print("该镜像没有写入钢印！")
    return True


def main_client_on_connect(client, userdata, flags, rc):
    log('Main client connected with result code ' + str(rc))

    log('Main client subscribe ' + updateCheckInTopic + ' topic')
    client.subscribe(updateCheckInTopic)

    log('Main client subscribe ' + updateResponseTopic + ' topic')
    client.subscribe(updateResponseTopic)

    log('Main client subscribe ' + cmd7ResTopic + ' topic')
    client.subscribe(cmd7ResTopic)


def main_client_on_message(client, userdata, msg):
    global updateList 
    global CheckPassFlag
    if msg.topic == updateCheckInTopic and CheckPassFlag == True:
        try:
            jsonMsg = json.loads(msg.payload.decode())

            if jsonMsg['gId'] not in ['20010801']:
                return

            updateDict = {}
            if 'version' in jsonMsg:
                if checkVersion(jsonMsg['version']):
                    updateDict['sn'] = jsonMsg['gId']
                    updateDict['package'] = 0
                    updateDict['time'] = time.time()
                    updateDict['send'] = True
                    updateList.append(updateDict)
            else:
                updateDict['sn'] = jsonMsg['gId']
                updateDict['package'] = 0
                updateDict['time'] = time.time()
                updateDict['send'] = True
                updateList.append(updateDict)
        except:
            log('Not a json data')

    elif msg.topic == updateResponseTopic and CheckPassFlag == True:
        try:
            responseMsg = msg.payload.decode().split()
            for each in updateList:
                if responseMsg[1] == each['sn']:
                    if responseMsg[2] == '0':
                        each['package'] = int(responseMsg[0]) + 1
                        each['time'] = time.time()
                        each['send'] = True
                    else:
                        each['send'] = True
        except:
            log('Error')
    elif msg.topic == cmd7ResTopic:
        try:
            Cmd7Res = json.loads(msg.payload)
            if 'stamping' in Cmd7Res:
                print("检查本地镜像")
                UpdataImageStamping = Last_Fourteen_Letter_Print(file_path)
                UpdataImageStamping = UpdataImageStamping.decode()
                if UpdataImageStamping == Cmd7Res['stamping']:
                    print("Romote Stamping {}".format(Cmd7Res['stamping']))
                    print("Local Stamping {}".format(UpdataImageStamping))
                    print("Equal")
                    CheckPassFlag = True
                else:
                    print("Romote Stamping {}".format(Cmd7Res['stamping']))
                    print("Local Stamping {}".format(UpdataImageStamping))
                    print("exit")
                    

        except:
            log("Error Cmd 7 Res")


def checkVersion(oldVersion):
    if 'Beta' in oldVersion and 'Beta' in tdlVersion or 'Beta' not in oldVersion and 'Beta' not in tdlVersion:
        if oldVersion < tdlVersion:
            return True
        else:
            return False
    elif 'Beta' in oldVersion and 'Beta' not in tdlVersion:
        if oldVersion[0:5] < tdlVersion:
            return True
        else:
            return False
    elif 'Beta' not in oldVersion and 'Beta' in tdlVersion:
        if oldVersion < tdlVersion[0:5]:
            return True
        else:                                                                                                
            return False
    return False


def createCheckCode(checkData):
    sumDat = 0
    codeBytes = b''
    for each in checkData:
        sumDat += each
    sumHex = hex(sumDat).lstrip('0x')
    while len(sumHex) != 8:
        sumHex = '0' + sumHex

    codeBytes += bytes((int(sumHex[0:2], 16),))
    codeBytes += bytes((int(sumHex[2:4], 16),))
    codeBytes += bytes((int(sumHex[4:6], 16),))
    codeBytes += bytes((int(sumHex[6:8], 16),))

    return codeBytes


def update():
    global fileData
    global updateList
    global mainClient
    global packageSize
    global oldTime

    size = get_Image_Size(old_filepath)
    data = [{"cmd":7,"LastImageSize":size}]
    data = json.dumps(data, separators=(',', ':'))
    # datas = json.dumps(datas)
    print(data)
    time.sleep(1)
    #  bytes((0,))   bytes((packageSize,))

    while CheckPassFlag == False:
        print("waiting for check Pass Flag...")
        mainClient.publish(cmd7Topic, data)
        time.sleep(5)

    curr_id = 0
    startPackage = curr_id.to_bytes(2,"big")  + b'Witium_Update' + packageSize.to_bytes(2,"big") 
    startPackage += createCheckCode(startPackage)
    while True:
        if len(updateList) == 0:
            time.sleep(1)
            
        else:
            curTime = time.time()
            if curTime - oldTime > 30:
                oldTime = time.time()
                log('Current List:' + str(updateList))

            for eachModule in updateList:
                if eachModule['send'] == True:
                    eachModule['send'] = False
                    if eachModule['package'] == 0:
                        mainClient.publish(updateTopic.format(eachModule['sn']), startPackage)
                    else:
                        if eachModule['package'] != packageSize:
                            sendData = eachModule['package'].to_bytes(2,"big") + fileData[(eachModule['package'] - 1) * 4096:eachModule['package'] * 4096]
                        else:
                            sendData = eachModule['package'].to_bytes(2,"big") + fileData[(eachModule['package'] - 1) * 4096:]
                        sendData += createCheckCode(sendData)
                        mainClient.publish(updateTopic.format(eachModule['sn']), sendData)
                    log('Send the ' + str(eachModule['package']) + ' package to ' + eachModule['sn'])
                    if eachModule['package'] == packageSize:
                        updateList.remove(eachModule)
                        log('Update success ' + eachModule['sn'])

                nowTime = time.time()
                if nowTime - eachModule['time'] > 30:
                    updateList.remove(eachModule)
                    log('Update time out ' + eachModule['sn'])


def log(logData):
    with open('log.txt', 'a+') as logFile:
        logFile.write(logData + '\n')
    print(logData)



#----------------------------------------------------------------------------------------#




path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
resourcesPath = path + os.sep + 'scripts' + os.sep
curFileList = os.listdir(os.getcwd())
oldTime = time.time()  

if not os.path.exists('log.txt'):
    with open('log.txt', 'w'):
        pass

file_path = resourcesPath + NewImage
old_filepath = resourcesPath + OldImage
print(file_path)

size = get_Image_Size(file_path)

print(f"新镜像文件大小：{size} bytes")
if Judge_Exit(Last_Fourteen_Letter_Print(file_path)) == True:
    print("即将退出...")
    time.sleep(2)
    exit(-1)
else:
    print("The stamping of Update Image is Pass")



log('file_path:' + file_path)
with open(file_path, 'rb') as file:
    fileData = file.read()
packageSize = (len(fileData) - 1) // 4096 + 1
log('packageSize:' + str(packageSize))

if not file_path:
    log('Update file has not found.')
else:
    mainClient = mqtt.Client()
    mainClient.username_pw_set(mqttConfig['userName'], mqttConfig['password'])
    mainClient.on_connect = main_client_on_connect
    mainClient.on_message = main_client_on_message
    mainClient.connect(mqttConfig['mqttServer'], mqttConfig['mqttPort'], 60)
    
    thread = threading.Thread(target=update)
    thread.setDaemon(True)
    thread.start()

    mainClient.loop_forever()




